@@ -18,8 +18,9 @@ module Yesod.Auth
       -- * Plugin interface
     , Creds (..)
     , setCreds
+    , clearCreds
       -- * User functions
-    , maybeAuthId
+    , defaultMaybeAuthId
     , maybeAuth
     , requireAuthId
     , requireAuth
@@ -59,41 +60,41 @@ type AuthRoute = Route Auth
 type Method = Text
 type Piece = Text
 
-data AuthPlugin m = AuthPlugin
+data AuthPlugin master = AuthPlugin
     { apName :: Text
-    , apDispatch :: Method -> [Piece] -> GHandler Auth m ()
-    , apLogin :: forall s. (Route Auth -> Route m) -> GWidget s m ()
+    , apDispatch :: Method -> [Piece] -> GHandler Auth master ()
+    , apLogin :: forall sub. (Route Auth -> Route master) -> GWidget sub master ()
     }
 
 getAuth :: a -> Auth
 getAuth = const Auth
 
 -- | User credentials
-data Creds m = Creds
+data Creds master = Creds
     { credsPlugin :: Text -- ^ How the user was authenticated
     , credsIdent :: Text -- ^ Identifier. Exact meaning depends on plugin.
     , credsExtra :: [(Text, Text)]
     }
 
-class (Yesod m, PathPiece (AuthId m), RenderMessage m FormMessage) => YesodAuth m where
-    type AuthId m
+class (Yesod master, PathPiece (AuthId master), RenderMessage master FormMessage) => YesodAuth master where
+    type AuthId master
 
     -- | Default destination on successful login, if no other
     -- destination exists.
-    loginDest :: m -> Route m
+    loginDest :: master -> Route master
 
     -- | Default destination on successful logout, if no other
     -- destination exists.
-    logoutDest :: m -> Route m
+    logoutDest :: master -> Route master
 
     -- | Determine the ID associated with the set of credentials.
-    getAuthId :: Creds m -> GHandler s m (Maybe (AuthId m))
+    getAuthId :: Creds master -> GHandler sub master (Maybe (AuthId master))
 
     -- | Which authentication backends to use.
-    authPlugins :: m -> [AuthPlugin m]
+    authPlugins :: master -> [AuthPlugin master]
 
     -- | What to show on the login page.
-    loginHandler :: GHandler Auth m RepHtml
+    loginHandler :: GHandler Auth master RepHtml
     loginHandler = defaultLayout $ do
         setTitleI Msg.LoginTitle
         tm <- lift getRouteToMaster
@@ -101,31 +102,57 @@ class (Yesod m, PathPiece (AuthId m), RenderMessage m FormMessage) => YesodAuth
         mapM_ (flip apLogin tm) (authPlugins master)
 
     -- | Used for i18n of messages provided by this package.
-    renderAuthMessage :: m
+    renderAuthMessage :: master
                       -> [Text] -- ^ languages
                       -> AuthMessage -> Text
     renderAuthMessage _ _ = defaultMessage
 
     -- | After login and logout, redirect to the referring page, instead of
     -- 'loginDest' and 'logoutDest'. Default is 'False'.
-    redirectToReferer :: m -> Bool
+    redirectToReferer :: master -> Bool
     redirectToReferer _ = False
 
     -- | Return an HTTP connection manager that is stored in the foundation
     -- type. This allows backends to reuse persistent connections. If none of
     -- the backends you're using use HTTP connections, you can safely return
     -- @error \"authHttpManager"@ here.
-    authHttpManager :: m -> Manager
+    authHttpManager :: master -> Manager
 
     -- | Called on a successful login. By default, calls
     -- @setMessageI NowLoggedIn@.
-    onLogin :: GHandler s m ()
+    onLogin :: GHandler sub master ()
     onLogin = setMessageI Msg.NowLoggedIn
 
     -- | Called on logout. By default, does nothing
-    onLogout :: GHandler s m ()
+    onLogout :: GHandler sub master ()
     onLogout = return ()
 
+    -- | Retrieves user credentials, if user is authenticated.
+    --
+    -- By default, this calls 'defaultMaybeAuthId' to get the user ID from the
+    -- session. This can be overridden to allow authentication via other means,
+    -- such as checking for a special token in a request header. This is
+    -- especially useful for creating an API to be accessed via some means
+    -- other than a browser.
+    --
+    -- Since 1.1.2
+    maybeAuthId :: GHandler sub master (Maybe (AuthId master))
+    maybeAuthId = defaultMaybeAuthId
+
+credsKey :: Text
+credsKey = "_ID"
+
+-- | Retrieves user credentials from the session, if user is authenticated.
+--
+-- Since 1.1.2
+defaultMaybeAuthId :: YesodAuth master
+                   => GHandler sub master (Maybe (AuthId master))
+defaultMaybeAuthId = do
+    ms <- lookupSession credsKey
+    case ms of
+        Nothing -> return Nothing
+        Just s -> return $ fromPathPiece s
+
 mkYesodSub "Auth"
     [ ClassP ''YesodAuth [VarT $ mkName "master"]
     ]
@@ -137,11 +164,11 @@ mkYesodSub "Auth"
 /page/#Text/STRINGS PluginR
 |]
 
-credsKey :: Text
-credsKey = "_ID"
-
--- | FIXME: won't show up till redirect
-setCreds :: YesodAuth m => Bool -> Creds m -> GHandler s m ()
+-- | Sets user credentials for the session after checking them with authentication backends.
+setCreds :: YesodAuth master
+         => Bool         -- ^ if HTTP redirects should be done
+         -> Creds master -- ^ new credentials
+         -> GHandler sub master ()
 setCreds doRedirects creds = do
     y    <- getYesod
     maid <- getAuthId creds
@@ -149,7 +176,10 @@ setCreds doRedirects creds = do
         Nothing ->
           when doRedirects $ do
             case authRoute y of
-              Nothing -> do rh <- defaultLayout $ toWidget [shamlet| <h1>Invalid login |]
+              Nothing -> do rh <- defaultLayout $ toWidget [shamlet|
+$newline never
+<h1>Invalid login
+|]
                             sendResponse rh
               Just ar -> do setMessageI Msg.InvalidLogin
                             redirect ar
@@ -159,7 +189,20 @@ setCreds doRedirects creds = do
               onLogin
               redirectUltDest $ loginDest y
 
-getCheckR :: YesodAuth m => GHandler Auth m RepHtmlJson
+-- | Clears current user credentials for the session.
+--
+-- Since 1.1.7
+clearCreds :: YesodAuth master
+           => Bool -- ^ if HTTP redirect to 'logoutDest' should be done
+           -> GHandler sub master ()
+clearCreds doRedirects = do
+    y <- getYesod
+    deleteSession credsKey
+    when doRedirects $ do
+        onLogout
+        redirectUltDest $ logoutDest y
+
+getCheckR :: YesodAuth master => GHandler Auth master RepHtmlJson
 getCheckR = do
     creds <- maybeAuthId
     defaultLayoutJson (do
@@ -168,6 +211,7 @@ getCheckR = do
   where
     html' creds =
         [shamlet|
+$newline never
 <h1>Authentication Status
 $maybe _ <- creds
     <p>Logged in.
@@ -181,23 +225,21 @@ $nothing
 
 setUltDestReferer' :: YesodAuth master => GHandler sub master ()
 setUltDestReferer' = do
-    m <- getYesod
-    when (redirectToReferer m) setUltDestReferer
+    master <- getYesod
+    when (redirectToReferer master) setUltDestReferer
 
-getLoginR :: YesodAuth m => GHandler Auth m RepHtml
+getLoginR :: YesodAuth master => GHandler Auth master RepHtml
 getLoginR = setUltDestReferer' >> loginHandler
 
-getLogoutR :: YesodAuth m => GHandler Auth m ()
-getLogoutR = setUltDestReferer' >> postLogoutR -- FIXME redirect to post
+getLogoutR :: YesodAuth master => GHandler Auth master ()
+getLogoutR = do
+    tm <- getRouteToMaster
+    setUltDestReferer' >> redirectToPost (tm LogoutR)
 
-postLogoutR :: YesodAuth m => GHandler Auth m ()
-postLogoutR = do
-    y <- getYesod
-    deleteSession credsKey
-    onLogout
-    redirectUltDest $ logoutDest y
+postLogoutR :: YesodAuth master => GHandler Auth master ()
+postLogoutR = clearCreds True
 
-handlePluginR :: YesodAuth m => Text -> [Text] -> GHandler Auth m ()
+handlePluginR :: YesodAuth master => Text -> [Text] -> GHandler Auth master ()
 handlePluginR plugin pieces = do
     master <- getYesod
     env <- waiRequest
@@ -206,41 +248,46 @@ handlePluginR plugin pieces = do
         [] -> notFound
         ap:_ -> apDispatch ap method pieces
 
--- | Retrieves user credentials, if user is authenticated.
-maybeAuthId :: YesodAuth m => GHandler s m (Maybe (AuthId m))
-maybeAuthId = do
-    ms <- lookupSession credsKey
-    case ms of
-        Nothing -> return Nothing
-        Just s -> return $ fromPathPiece s
-
-maybeAuth :: ( YesodAuth m
-             , b ~ YesodPersistBackend m
+maybeAuth :: ( YesodAuth master
+#if MIN_VERSION_persistent(1, 1, 0)
+             , PersistMonadBackend (b (GHandler sub master)) ~ PersistEntityBackend val
+             , b ~ YesodPersistBackend master
+             , Key val ~ AuthId master
+             , PersistStore (b (GHandler sub master))
+#else
+             , b ~ YesodPersistBackend master
              , b ~ PersistEntityBackend val
-             , Key b val ~ AuthId m
-             , PersistStore b (GHandler s m)
+             , Key b val ~ AuthId master
+             , PersistStore b (GHandler sub master)
+#endif
              , PersistEntity val
-             , YesodPersist m
-             ) => GHandler s m (Maybe (Entity val))
+             , YesodPersist master
+             ) => GHandler sub master (Maybe (Entity val))
 maybeAuth = runMaybeT $ do
     aid <- MaybeT $ maybeAuthId
     a   <- MaybeT $ runDB $ get aid
     return $ Entity aid a
 
-requireAuthId :: YesodAuth m => GHandler s m (AuthId m)
+requireAuthId :: YesodAuth master => GHandler sub master (AuthId master)
 requireAuthId = maybeAuthId >>= maybe redirectLogin return
 
-requireAuth :: ( YesodAuth m
-               , b ~ YesodPersistBackend m
+requireAuth :: ( YesodAuth master
+               , b ~ YesodPersistBackend master
+#if MIN_VERSION_persistent(1, 1, 0)
+               , PersistMonadBackend (b (GHandler sub master)) ~ PersistEntityBackend val
+               , Key val ~ AuthId master
+               , PersistStore (b (GHandler sub master))
+#else
                , b ~ PersistEntityBackend val
-               , Key b val ~ AuthId m
-               , PersistStore b (GHandler s m)
+               , Key b val ~ AuthId master
+               , PersistStore b (GHandler sub master)
+#endif
                , PersistEntity val
-               , YesodPersist m
-               ) => GHandler s m (Entity val)
+               , YesodPersist master
+               ) => GHandler sub master (Entity val)
 requireAuth = maybeAuth >>= maybe redirectLogin return
 
-redirectLogin :: Yesod m => GHandler s m a
+redirectLogin :: Yesod master => GHandler sub master a
 redirectLogin = do
     y <- getYesod
     setUltDestCurrent
@@ -248,7 +295,7 @@ redirectLogin = do
         Just z -> redirect z
         Nothing -> permissionDenied "Please configure authRoute"
 
-instance YesodAuth m => RenderMessage m AuthMessage where
+instance YesodAuth master => RenderMessage master AuthMessage where
     renderMessage = renderAuthMessage
 
 data AuthException = InvalidBrowserIDAssertion
